; File:		tn12-isp-simulator\main.asm
; Device:	AVR
; Created:	2025-01-28
; Version:	2025-01-31
; Author:	Johannes Fechner

; Devices to be supported (in future):
.equ T3227 = 0 ; not supported yet
.equ A64DD32 = 1

; Device selection:
.equ MCU = A64DD32

.nolist
.if MCU == A64DD32
.include "C:\Program Files\Microchip\MPLABX\v6.20\packs\Microchip\AVR-Dx_DFP\2.4.286\avrasm\inc\AVR64DD32def.inc"
.else
.error "Device not supported yet."
.endif
.include "macros.inc"
.list

.equ SIMFLASH_SIZE = 1024

; CPU clock frequency:
.equ F_CPU = 20_000_000 ; Hz

; SPI for communication with host/"programmer":
.if MCU == T3227
.equ SPI_VDIR = VPORTA_DIR
.equ SPI_MISO_bp = 2
.equ SPI_SS_bp = 4
.equ SPI_SS_PINCTRL = PORTA_PIN4CTRL

.elif MCU == A64DD32
.equ SPI_VDIR = VPORTA_DIR
.equ SPI_MISO_bp = 5
.equ SPI_SS_bp = 7
.equ SPI_SS_PINCTRL = PORTA_PIN7CTRL

; UART for sending debugging messages:
.equ UART_DBG_BPS = 115_200 ; bits/s
.equ UART_DBG_VPORT_DIR = VPORTA_DIR
.equ UART_DBG_TXD_bp = 0

.else
.error "Device not supported yet."
.endif

; == Register name definitions ==
.def zero = r2		; Contains always 0x00.
.def srCopy = r3
.def wri0Copy = r4
.def wri1Copy = r5
.def ZLCopy = r14
.def ZHCopy = r15
.def wri0 = r16
.def wri1 = r17
.def wri2 = r18
.def wri3 = r19
.def wri4 = r20
.def spiIsr_dataCntr = r21
.def wriL = r24
.def wriH = r25

; == Data memory usage ==
.dseg
.org SRAM_START
spiData_sram: .byte 4
simFlash_sram: .byte SIMFLASH_SIZE

.cseg
.org 0
	rjmp	reset

; == Interrupt vectors ==
.org PORTA_PORT_vect
	rjmp	spiSsIsr
.org TCA0_OVF_vect
	rjmp	tcaOvfIsr
.org SPI0_INT_vect
	rjmp	spiIsr

.org INT_VECTORS_SIZE

; == Constant data ==
signature_flash: .db 0x1E, 0x90, 0x05, 0
msgSsFallingEdge_flash: .db "Executing SPI SS falling edge ISR", '\r', '\n', 0, 0
msgChipErase_flash: .db "Chip Erase", '\r', '\n', 0
msgReadPgmMem_flash: .db "Read Program Memory, byte address ", 0
msgData_flash: .db ", data ", 0
msgWritePgmMem_flash: .db "Write Program Memory, byte addr ", 0
msgReadSignature_flash: .db "Read Signature Byte ", 0
msgLineBreak_flash: .db '\r', '\n', 0, 0

; == Execution entry point after reset ==
reset:

; == Initialize registers ==
	clr		zero

; == Initialize SPI_SS pin ==
	ldi		wri0, PORT_PULLUPEN_bm | PORT_ISC_FALLING_gc
	xout	SPI_SS_PINCTRL, wri0

.if MCU == T3227
; == Disable clock prescaler ==
	ldi		wri0, CPU_CCP_IOREG_gc
	xout	CPU_CCP, wri0
	xin		wri0, CLKCTRL_MCLKCTRLB
	andi	wri0, ~CLKCTRL_PEN_bm
	xout	CLKCTRL_MCLKCTRLB, wri0
	
.elif MCU == A64DD32
; == Select OSCHF frequency ==
	ldi		wri0, CPU_CCP_IOREG_gc
	xout	CPU_CCP, wri0
	ldi		wri0, CLKCTRL_FRQSEL_20M_gc
	xout	CLKCTRL_OSCHFCTRLA, wri0
.endif

; == Initialize SPI ==
; Connected with host/"programmer" MCU.
; Client role.
	ldi		wri0, SPI_IE_bm
	xout	SPI0_INTCTRL, wri0
	ldi		wri0, SPI_ENABLE_bm
	xout	SPI0_CTRLA, wri0
	sbi		SPI_VDIR, SPI_MISO_bp
	
; == Initialize TCA0 ==
; 100-us periodic interrupt.
.equ TCA0_PERIOD = ROUND(F_CPU*1.0e-4)
	ldi		wri0, low(TCA0_PERIOD)
	xout	TCA0_SINGLE_PERL, wri0
	ldi		wri0, high(TCA0_PERIOD)
	xout	TCA0_SINGLE_PERH, wri0
	ldi		wri0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTCTRL, wri0

; == Initialize USART0 ==
; Debug message output.
.equ BAUD0 = ROUND(4.0*F_CPU/(1.0*UART_DBG_BPS))
	ldi		wri0, low(BAUD0)
	xout	USART0_BAUDL, wri0
	ldi		wri0, high(BAUD0)
	xout	USART0_BAUDH, wri0
	ldi		wri0, USART_TXEN_bm
	xout	USART0_CTRLB, wri0
	sbi		UART_DBG_VPORT_DIR, UART_DBG_TXD_bp
	
; == Enable interrupts ==
	sei
	
loop:
	rjmp	loop

; == SPI SS Falling-Edge Interrupt Routine ==
spiSsIsr:
	cli
; Register saving currently omitted
; because main loop is empty.
; Clear interrupt flag:
	ldi		wri0, 1<<SPI_SS_bp
	xout	PORTA_INTFLAGS, wri0
; Send debug message:
	ldiz	2*msgSsFallingEdge_flash
	rcall	sendDbgMsg
; Clear the SPI data byte counter:
	clr		spiIsr_dataCntr
	sei
	reti

; == TCA0 Overflow Interrupt Routine ==
tcaOvfIsr:
	reti

; == SPI Interrupt Routine ==
spiIsr:
	cli
; Register saving currently omitted
; because main loop is empty.
; Clear interrupt flag:
	ldi		wri0, SPI_IF_bm
	xout	SPI0_INTFLAGS, wri0
; Get the received byte and store it:
	xin		wri0, SPI0_DATA
	cpi		spiIsr_dataCntr, 4
	brlo	spiIsr2 ; OK
	clr		spiIsr_dataCntr
	rjmp	spiIsr0
spiIsr2:
	ldiz	spiData_sram
	addz	spiIsr_dataCntr
	st		Z, wri0
	inc		spiIsr_dataCntr
	cpi		spiIsr_dataCntr, 3
	breq	spiIsr_chipErase
	rjmp	spiIsr3
; Three data bytes received.

spiIsr_chipErase:
; == Test for Chip Erase command ==
	lds		wri0, spiData_sram ; first byte
	cpi		wri0, 0xAC
	brne	spiIsr1
	lds		wri0, spiData_sram+1 ; second byte
	andi	wri0, 0b1110_0000
	cpi		wri0, 0b1000_0000
	brne	spiIsr1
; Chip Erase command received.
	ldiz	2*msgChipErase_flash
	rcall	sendDbgMsg
	ldiz	simFlash_sram
	ldi		wriL, low(SIMFLASH_SIZE)
	ldi		wriH, high(SIMFLASH_SIZE)
spiIsr4:
	st		Z+, zero
	sbiw	wriL, 1
	brne	spiIsr4
	rjmp	spiIsr0
	
spiIsr1:
; == Test for Read Program Memory command ==
	lds		wri0, spiData_sram ; first byte
	andi	wri0, 0b1111_0111
	cpi		wri0, 0b0010_0000
	brne	spiIsr5
; Read Program Memory command received.
	ldiz	2*msgReadPgmMem_flash
	rcall	sendDbgMsg
	lds		wri0, spiData_sram+1 ; second byte
	andi	wri0, 0x01
	mov		YH, wri0
	lds		wri0, spiData_sram+2 ; third byte
	mov		YL, wri0
; Convert the received word address to byte address:
	lsl		YL
	rol		YH
	lds		wri0, spiData_sram ; command byte, bit #3 selects low/high byte
	sbrc	wri0, 3
	ori		YL, 1
	mov		wri0, YH
	rcall	sendDbgHex
	mov		wri0, YL
	rcall	sendDbgHex
; Load the addressed byte and move it into SPI transmit register:
	ld		wri0, Y
	xout	SPI0_DATA, wri0
	ldiz	2*msgData_flash
	rcall	sendDbgMsg
	rcall	sendDbgHex
	ldiz	2*msgLineBreak_flash
	rcall	sendDbgMsg
	rjmp	spiIsr0
	
spiIsr5:
; == Test for Read Signature Byte command ==
	lds		wri0, spiData_sram ; first byte
	cpi		wri0, 0x30
	brne	spiIsr0
; Read Signature Byte command received.
	ldiz	2*msgReadSignature_flash
	rcall	sendDbgMsg
	lds		wri0, spiData_sram+2 ; third byte
	rcall	sendDbgHex
	ldiz	2*msgLineBreak_flash
	rcall	sendDbgMsg
	ldiz	2*signature_flash
	addz	wri0
	lpm		wri0, Z
	xout	SPI0_DATA, wri0
	rjmp	spiIsr0
	
spiIsr3:
	cpi		spiIsr_dataCntr, 4
	brne	spiIsr0
; Four data bytes received.
	clr		spiIsr_dataCntr

; == Test for Write Program Memory command ==
	lds		wri0, spiData_sram ; first byte
	andi	wri0, 0b1111_0111
	cpi		wri0, 0b0100_0000
	brne	spiIsr0
; Write Program Memory command received.
	ldiz	2*msgWritePgmMem_flash
	rcall	sendDbgMsg
	lds		wri0, spiData_sram+1 ; second byte
	andi	wri0, 0x01
	mov		YH, wri0
	lds		wri0, spiData_sram+2 ; third byte
	mov		YL, wri0
; Convert the received word address to byte address:
	lsl		YL
	rol		YH
	lds		wri0, spiData_sram ; command byte, bit #3 selects low/high byte
	sbrc	wri0, 3
	ori		YL, 1
	mov		wri0, YH
	rcall	sendDbgHex
	mov		wri0, YL
	rcall	sendDbgHex
; Store the last input byte to the obtained address:
	lds		wri0, spiData_sram+3 ; fourth byte
	st		Y, wri0
	ldiz	2*msgData_flash
	rcall	sendDbgMsg
	rcall	sendDbgHex
	ldiz	2*msgLineBreak_flash
	rcall	sendDbgMsg
	
spiIsr0:
	sei
	reti

; == Routine: Wait for previous USART0 transmission to finish and send next byte ==
; Parameter: wri0, the byte to be sent.
sendDbgByte:
	push	wri0
; Make sure the transmit data registers are empty:
sendDbgByte0:
	xin		wri0, USART0_STATUS
	sbrs	wri0, USART_DREIF_bp
	rjmp	sendDbgByte0
	pop		wri0
	xout	USART0_TXDATAL, wri0
	ret

; === Routine: Send string via USART0 ===
; Parameter: Z pointer, start address of the null-terminated string.
; String length is limited to 255 characters.
sendDbgMsg:
	push	wri0
	push	wri1
	clr		wri1
sendDbgMsg_loop:
	lpm		wri0, Z+
	tst		wri0
	breq	sendDbgMsg_end
	rcall	sendDbgByte
	inc		wri1
	cpi		wri1, $FF
	breq	sendDbgMsg_end
	rjmp	sendDbgMsg_loop
sendDbgMsg_end:
	pop		wri1
	pop		wri0
	ret

; === Routine: Send byte in hexadecimal via USART0 ===
; Send 8-bit unsigned number as two hex digits.
; Parameter: wri0.
sendDbgHex:
	push	wri0
	swap	wri0
	andi	wri0, 0x0F
	rcall	sendDbgHexDigit
	pop		wri0
	push	wri0
	andi	wri0, 0x0F
	rcall	sendDbgHexDigit
	pop		wri0
	ret
sendDbgHexDigit:
	cpi		wri0, 10
	brlo	sendDbgHexDigit0
	subi	wri0, -('A'-'9'-1)
sendDbgHexDigit0:
	subi	wri0, -'0'
	rcall	sendDbgByte
	ret
